﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.ComponentModel;

namespace SpringbearWorks.WinForms.Lcd
{
    /*
     * History
     * ===========
     * Greg H.C.
     * May 5th, 2012
     * Changed namespace to SpringbearWorks.WinForms.Lcd.
     * 
     * May 4th, 2012
     * Reorgnized source code.
     * 
     * May 3rd, 2012
     * Finished with DotMatrixCharacter class.
     * Expanded LcdCharset.Charset with Chinese words and symbols from 0x80 to
     * 0x9b.
     * See page 20 of VP. LCM Datasheet
     * 
     * Design Note
     * ===========
     * Greg H.C.
     * May 2nd, 2012
     * 
     * This design made each character a control. This is easy to understand
     * and to implement, but VERY BAD IN PERFORMANCE. Especially if works in
     * a mass size of LCM (16x2 or larger). Blinks can be seen apparently.
     * 
     * Consider to solve this problem by not derive from Windows.Forms.Control,
     * and make a GetImage() method to report the final appearance of a char
     * to its owner (LCM screen).
     * 
     * */
    [ToolboxBitmap("Lcd\\lcd_char_logo.bmp")]
    public class DotMatrixCharacter : Control
    {
        #region Constants
        // a fixed 5x8 dot matrix
        const int DOT_COLS = 5;
        const int DOT_ROWS = 8;
        #endregion

        #region Variables
        SolidBrush activeBrush;
        SolidBrush inactiveBrush;
        //SolidBrush backBrush;

        // character to display, look up in LcdCharset.Charset (256 bytes)
        byte ddramAddress = 0;

        // custom CGRAM data
        byte[] cgramData;

        int spacing = 1;
        SizeF blockSize = new SizeF();

        // use dot blocks
        bool circleBlock = false;
        // display source (Auto, DDRAM, CGRAM)
        DisplaySource charSource;
        #endregion

        #region Properties
        /// <summary>
        /// Show circle dots instead of rectangle ones.
        /// </summary>
        [Browsable(true)]
        public bool CircleBlock
        {
            get { return this.circleBlock; }
            set
            {
                this.circleBlock = value;
                this.Invalidate();
            }
        }

        [Browsable(true)]
        /// <summary>
        /// Get or set 8 bytes of data of a CGRAM character
        /// </summary>
        public byte[] Cgram
        {
            get { return cgramData; }
            set
            {
                if (value == null || value.Length != DOT_ROWS)
                    cgramData = null;
                else
                    cgramData = value;

                // refresh
                this.Invalidate();
            }
        }

        /// <summary>
        /// Get or set address of DDRAM.
        /// </summary>
        public byte Ddram
        {
            get { return this.ddramAddress; }
            set
            {
                this.ddramAddress = value;
                this.Invalidate();
            }
        }

        /// <summary>
        /// Get or set display source between Auto, CGRAM and DDRAM.
        /// </summary>
        public DisplaySource DisplaySource
        {
            get { return this.charSource; }
            set
            {
                this.charSource = value;
                this.Invalidate();
            }
        }

        [DefaultValue(0)]
        /// <summary>
        /// Get or set address of DDRAM from a character.
        /// </summary>
        public char Character
        {
            get { return (char)this.ddramAddress; }
            set
            {
                this.ddramAddress = (byte)value;
                this.Invalidate();
            }
        }

        [DefaultValue(1), Description("Spacing between blocks")]
        /// <summary>
        /// Spacing between blocks in pixel.
        /// </summary>
        public int Spacing
        {
            get { return spacing; }
            set
            {
                if (value >= 0 && value < 10)
                {
                    spacing = value;
                    // make a new resize
                    OnResize(null);
                    this.Invalidate();
                }
            }
        }

        [DefaultValue(typeof(Color), "Color.Black"),
        Description("Active block color")]
        public override Color ForeColor
        {
            get { return activeBrush.Color; }
            set
            {
                activeBrush.Color = value;
                this.Invalidate();
            }
        }

        [DefaultValue(typeof(Color), "Color.YellowGreen"),
        Description("Inactive block color")]
        public Color InactiveColor
        {
            get { return inactiveBrush.Color; }
            set
            {
                inactiveBrush.Color = value;
                this.Invalidate();
            }
        }

        #endregion

        #region Initiates
        public DotMatrixCharacter()
        {
            activeBrush = new SolidBrush(Color.Black);
            inactiveBrush = new SolidBrush(Color.YellowGreen);
            //backBrush = new SolidBrush(Color.GreenYellow);

            //cgramData = new byte[DOT_ROWS];

            // set default size (height is auto)
            this.Width = 20;

            // forget Auto, use DDRAM & CGRAM
            this.charSource = DisplaySource.DDRAM;

            // clear cgram
            this.cgramData = null;

            this.DoubleBuffered = true;
            this.SetStyle(
                ControlStyles.OptimizedDoubleBuffer |
                ControlStyles.ResizeRedraw |
                ControlStyles.UserPaint,
                true
                );
        }
        #endregion

        #region Paints
        void drawBlocks(Graphics g)
        {
            byte[] charData;

            // check source of char to display
            // for CGRAM support
            switch (charSource)
            {
                case DisplaySource.CGRAM:
                    if (cgramData == null || cgramData.Length != DOT_ROWS)
                        // invalid data, draw empty
                        // all 0x00
                        charData = new byte[DOT_ROWS];
                    else
                        charData = cgramData;
                    break;
                case DisplaySource.DDRAM:
                default:
                    //charData = CharacterGenerator.charset[ddramAddress];
                    charData = CharacterGenerator.GetDdram(ddramAddress);
                    break;
                //case DisplaySource.Auto:
                //default:
                //    // default is auto
                //    if (cgramData == null || cgramData.Length != DOT_ROWS)
                //        //charData = CharacterGenerator.charset[ddramAddress];
                //        charData = CharacterGenerator.GetDdram(ddramAddress);
                //    else
                //        charData = cgramData;
                //    break;
            }


            // ready to draw
            byte mask;

            for (int i = 0; i < DOT_ROWS; i++)
            {
                // if use mask = 0x01 (right to left)
                // the output will be vertical mirrored
                mask = 0x01 << (DOT_COLS - 1);

                for (int j = 0; j < DOT_COLS; j++)
                {
                    if ((mask & charData[i]) == 0)
                    {
                        // 0 - empty
                        if (circleBlock)
                            g.FillEllipse(
                                inactiveBrush,
                                j * (blockSize.Width + spacing),
                                i * (blockSize.Height + spacing),
                                blockSize.Width,
                                blockSize.Height
                                );
                        else
                            g.FillRectangle(
                                inactiveBrush,
                                j * (blockSize.Width + spacing),
                                i * (blockSize.Height + spacing),
                                blockSize.Width,
                                blockSize.Height
                                );
                    }
                    else
                    {
                        // 1 - fill
                        if (circleBlock)
                            g.FillEllipse(
                                activeBrush,
                                j * (blockSize.Width + spacing),
                                i * (blockSize.Height + spacing),
                                blockSize.Width,
                                blockSize.Height
                                );
                        else
                            g.FillRectangle(
                                activeBrush,
                                j * (blockSize.Width + spacing),
                                i * (blockSize.Height + spacing),
                                blockSize.Width,
                                blockSize.Height
                            );
                    }

                    // next bit
                    //mask <<= 1;
                    // msb to lsb
                    mask >>= 1;
                }
            }
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            drawBlocks(e.Graphics);
        }
        #endregion

        #region Sizing
        protected override void OnResize(EventArgs e)
        {
            // square blocks
            this.blockSize.Width = (float)(this.Width - spacing * (DOT_COLS - 1)) / (float)DOT_COLS;
            this.blockSize.Height = this.blockSize.Width;

            //this.Height = this.Width * DOT_ROWS / DOT_COLS;
            this.Height = (int)(spacing * (DOT_ROWS - 1) + this.blockSize.Height * DOT_ROWS);
        }
        #endregion
    }

    /// <summary>
    /// Character display source
    /// </summary>
    public enum DisplaySource
    {
        //Auto = 0,
        DDRAM = 1,
        CGRAM = 2,
    }
}
