﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace SpringbearWorks.WinForms.Lcd
{
    /*
     * Bugs
     * ===========
     * Greg H.C.
     * #002 Solved
     * May 5th, 2012
     * AutoSize miss functional.
     * 
     * #001 Serious
     * May 5th, 2012
     * All custom roperties must be set at runtime.
     * (Design time sets take no effect)??? What the f...
     * 
     * 
     * History
     * ===========
     * Greg H.C.
     * May 5th, 2012
     * Added "InactiveColor" property.
     * */

    [ToolboxBitmap("Lcd\\lcd_logo.bmp")]
    public partial class DotMatrixLcd : UserControl
    {
        #region Initiates
        public DotMatrixLcd()
        {
            InitializeComponent();
        }
        #endregion

        #region Constants
        const int warningThreshold = 40;
        #endregion

        #region Variables
        int rows = 2;
        int cols = 16;

        int charWidth = 20;

        DotMatrixCharacter[][] characters;

        string dispText = "";

        #endregion

        #region Properties

        [Browsable(true)]
        public override string Text
        {
            get { return this.dispText; }
            set
            {
                // text
                this.dispText = value;
                // NOTE: on hardware, NO clear is taken before write new
                //Clear();
                setText(this.dispText);
            }
        }

        private Color inactiveColor = Color.YellowGreen;
        public Color InactiveColor
        {
            get { return this.inactiveColor; }
            set
            {
                this.inactiveColor = value;
                if (characters != null)
                {
                    for (int i = 0; i < rows; i++)
                    {
                        for (int j = 0; j < cols; j++)
                        {
                            characters[i][j].InactiveColor = this.inactiveColor;
                        }
                    }
                }
            }
        }
        #endregion

        #region Methods
        public void Clear()
        {
            if (characters == null)
                return;
            for (int i = 0; i < this.rows; i++)
            {
                for (int j = 0; j < this.cols; j++)
                {
                    characters[i][j].Ddram = 0x00;
                }
            }
        }
        void setText(string text)
        {
            if (characters == null)
                return;
            byte[] tmp = Encoding.ASCII.GetBytes(text);

            for (int i = 0; (i < tmp.Length) && (i < this.rows * this.cols); i++)
            {
                characters[i / this.cols][i % this.cols].Ddram = tmp[i];
            }
        }
        #endregion

        #region Indexers

        /// <summary>
        /// returns DDRAM address
        /// </summary>
        /// <param name="row"></param>
        /// <param name="col"></param>
        /// <returns></returns>
        public byte this[int row, int col]
        {
            get { return characters[row][col].Ddram; }
            set { characters[row][col].Ddram = value; }
        }

        public void SetCharacter(byte value, int row, int col)
        {
            if (row < 0 || row >= this.rows)
                return;
            if (col < 0 || col >= this.cols)
                return;

            characters[row][col].Ddram = value;
        }

        public void SetCGRAM(byte[] value, int row, int col)
        {
            if (row < 0 || row >= this.rows)
                return;
            if (col < 0 || col >= this.cols)
                return;

            characters[row][col].Cgram = value;
        }

        public void EnableCGRAM(bool value, int row, int col)
        {
            if (row >= 0 && col >= 0)
            {
                if (value)
                    // use cgram
                    characters[row][col].DisplaySource = DisplaySource.CGRAM;
                else
                    // use ddram (not auto mode to reserve previous CGRAM data)
                    characters[row][col].DisplaySource = DisplaySource.DDRAM;
            }
        }

        #endregion

        #region Sizing
        public override bool AutoSize
        {
            get { return base.AutoSize; }
            set
            {
                base.AutoSize = value;
                OnResize(null);
            }
        }

        public int CharacterWidth
        {
            get { return this.charWidth; }
            set
            {
                if (value > 0 && characters != null)
                {
                    this.charWidth = value;
                    this.Controls.Clear();
                    for (int i = 0; i < rows; i++)
                    {
                        for (int j = 0; j < cols; j++)
                        {
                            characters[i][j].Width = this.charWidth;
                            characters[i][j].Location = new Point(
                                j * (characters[i][j].Width + this.Margin.All),
                                i * (characters[i][j].Height + this.Margin.All)
                                );
                        }
                        this.Controls.AddRange(characters[i]);
                    }

                    OnResize(null);
                }
            }
        }

        public int Rows
        {
            get { return this.rows; }
            set
            {
                if (value > 0)
                {
                    if (value > warningThreshold)
                    {
                        if (MessageBox.Show(
                            string.Format("Create a {0}x{1} LCD\nIt'll take minutes, continue?", this.cols, value),
                            "Warning",
                            MessageBoxButtons.YesNo,
                            MessageBoxIcon.Warning
                            ) == DialogResult.No)
                            return;
                    }

                    this.rows = value;
                    //characters = createCharacters(this.rows, this.cols);
                    createCharacters(this.rows, this.cols);

                    setText(this.dispText);
                    // resize
                    resizeRowColumn(this.rows, this.cols);
                }
            }
        }

        public int Columns
        {
            get { return this.cols; }
            set
            {
                if (value > 0)
                {
                    if (value > warningThreshold)
                    {
                        if (MessageBox.Show(
                            string.Format("Create a {0}x{1} LCD\nIt'll take minutes, continue?", value, this.rows),
                            "Warning",
                            MessageBoxButtons.YesNo,
                            MessageBoxIcon.Warning
                            ) == DialogResult.No)
                            return;
                    }
                    this.cols = value;
                    //characters = createCharacters(this.rows, this.cols);
                    createCharacters(this.rows, this.cols);

                    setText(this.dispText);
                    resizeRowColumn(this.rows, this.cols);
                }
            }
        }

        protected override void OnResize(EventArgs e)
        {
            if (characters != null && this.AutoSize)
            {
                this.AutoSizeMode = AutoSizeMode.GrowAndShrink;
            }
        }

        private void resizeRowColumn(int row, int col)
        {
            if (characters == null)
                return;

            if (row > 0 && col > 0)
            {
                this.Controls.Clear();
                // arrange array
                for (int i = 0; i < row; i++)
                {
                    for (int j = 0; j < col; j++)
                    {
                        characters[i][j].Location = new Point(
                            j * (characters[i][j].Width + this.Margin.All),
                            i * (characters[i][j].Height + this.Margin.All)
                            );
                    }
                    this.Controls.AddRange(characters[i]);
                }
            }
        }

        protected override void OnMarginChanged(EventArgs e)
        {
            if (characters != null)
            {
                resizeRowColumn(this.rows, this.cols);
            }
        }

        protected override void OnPaddingChanged(EventArgs e)
        {
            if (characters != null)
            {
                for (int i = 0; i < this.rows; i++)
                {
                    for (int j = 0; j < this.cols; j++)
                    {
                        characters[i][j].Spacing = this.Padding.All;
                    }
                }

                resizeRowColumn(this.rows, this.cols);
            }
        }

        //private DotMatrixCharacter[][] createCharacters(int row, int col)
        private void createCharacters(int row, int col)
        {
            //DotMatrixCharacter[][] tmp = new DotMatrixCharacter[row][];
            Array.Resize(ref characters, row);
            for (int i = 0; i < row; i++)
            {
                //tmp[i] = new DotMatrixCharacter[col];
                Array.Resize(ref characters[i], col);
                for (int j = 0; j < col; j++)
                {
                    characters[i][j] = new DotMatrixCharacter();
                    characters[i][j].Spacing = this.Padding.All;
                    characters[i][j].Cgram = null;
                    characters[i][j].Width = this.charWidth;
                }
            }

            //return tmp;


        }


        #endregion

        #region Appearance
        protected override void OnPaint(PaintEventArgs e)
        {
            foreach (Control c in this.Controls)
            {
                c.Invalidate();
            }
        }
        #endregion

        #region Paints
        private void DotMatrixLcd_Paint(object sender, PaintEventArgs e)
        {
            //foreach (Control c in this.Controls)
            //{
            //    c.Invalidate();
            //}
        }
        #endregion
    }
}
